use reference::Referencable;
use serde_json::{Map, Value};
use valid::Validate;

use crate::{col, get_mongo, id_of};

pub mod historic;
pub mod reference;
pub mod update;
pub mod valid;

/// Error type when updating a model
#[derive(Debug)]
pub enum UpdateError {
    /// Provided data was no object
    NoObject,
    /// Database related error
    Database,
    /// Validation failed
    Validation,
}

pub trait Model:
    Sized + Referencable + Validate + serde::Serialize + for<'a> serde::Deserialize<'a>
{
    /// Insert the model into the database
    async fn insert(&self) {
        let db = get_mongo!();
        let collection = col!(db, Self::collection_name());
        collection
            .insert_one(mongodb::bson::to_document(self).unwrap(), None)
            .await
            .unwrap();
    }

    /// Get a model by id from the database
    async fn get(id: &str) -> Option<Self> {
        let db = get_mongo!();
        let collection = col!(db, Self::collection_name());
        let doc = collection.find_one(id_of!(id), None).await.ok()??;
        mongodb::bson::from_document(doc).ok()
    }

    /// Update values of model into database
    async fn update(&mut self, data: &serde_json::Value) -> Result<(), UpdateError> {
        // get db collection
        let db = get_mongo!();
        let collection = col!(db, Self::collection_name());
        let mut update = mongodb::bson::Document::new();

        if let Some(obj) = data.as_object() {
            // run model specific update
            self.update_values(obj, &mut update).await;

            // validate and update
            if self.validate().await {
                collection
                    .update_one(
                        id_of!(self.id()),
                        mongodb::bson::doc! {"$set": update },
                        None,
                    )
                    .await
                    .map_err(|_| UpdateError::Database)?;
                return Ok(());
            } else {
                return Err(UpdateError::Validation);
            }
        }

        Err(UpdateError::NoObject)
    }

    /// Update the `Model` based on the provided JSON Object.
    ///
    /// This function should be implemented for a `Model`. It should update all keys from the JSON Object.
    /// For every updated value in the JSON Object the `Model`s fields should be updated and the value should be put in the `update` `Document`
    ///
    /// To avoid making mistakes in the Update logic use the macros from `update.rs`
    async fn update_values(
        &mut self,
        obj: &Map<String, Value>,
        update: &mut mongodb::bson::Document,
    );
}
